/* $Id: common-run.cpp 411 2010-05-06 19:29:33Z oh2gve $
 *
 * Copyright 2007, 2008, 2009, 2010 Tapio Aaltonen
 *
 *
 * This file is part of aprsg.
 *
 * aprsg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * aprsg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with aprsg.  If not, see <http://www.gnu.org/licenses/>.
 *
 *
 * APRS is a registered trademark of APRS Software and Bob Bruninga, WB4APR.
*/

// Own stuff.
#ifdef __WXMSW__
#	include "../windows/controller.h"
#else
#	include "main.h"
#endif
#include "cfg.h"
#include "packetlog.h"
#include "dupefilter.h"
#include "guard.h"
#include "beacon.h"

// Std stuff.
#include <cstdlib>
#include <vector>
using std::vector;
#include <string>
using std::string;
#ifndef __WXMSW__
#	include <csignal>
#endif

// Wx stuff.
#include <wx/log.h>



#ifdef __WXMSW__
wxThread::ExitCode Controller::Entry()
#else
int AprsApp::OnRun()
#endif
{
#ifndef __WXMSW__
	if ( iAmDaddy_ )
	{
		return EXIT_SUCCESS;
	}
#else
	wxLogMessage(wxT("Controller started."));
#endif

	// Initialize packet logging.
	Packetlog packetlog(configuration_.logcfg, configuration_.callsign, configuration_.location);
	if ( !packetlog )
	{
#ifdef __WXMSW__
		eventToMain(ID_fatalError);
		return NULL;
#else
		return EXIT_FAILURE;
#endif
	}

	// Initialize duplicate filtering.
	DupeFilter dupefilter(configuration_.duplicateTime);

	// Initialize guards for RF connections.
	vector<Guard*> guards;
	for ( unsigned int i = 0; i < configuration_.connections.size(); ++i )
	{
		guards.push_back( new Guard(connections_.at(i), configuration_.connections.at(i), i+1, configuration_.connections.size()+1, dupefilter, packetlog, string(configuration_.callsign.ToAscii()) ) );
	}

	// Initialize beaconing.
	Beacon beacon(connections_, uplink_, configuration_, dupefilter, packetlog);
	if ( !beacon )
	{
		wxLogMessage(wxString::FromAscii(beacon.error().c_str()));
		for ( unsigned int i = 0; i < guards.size(); ++i )
		{
			delete guards.at(i);
		}
#ifdef __WXMSW__
		eventToMain(ID_fatalError);
		return NULL;
#else
		return EXIT_FAILURE;
#endif
	}

	// Begin gateway action.
	wxLogMessage(wxT("Startup successfull, entering operation mode."));
	bool noError = true;

	// Counts time in 10 ms chunks.
	unsigned int maintCounter = 0;


	// Here is THE main loop.
#ifdef __WXMSW__
	while ( noError && !TestDestroy() )
#else
	while ( noError )
#endif
	{
		// Check guards for input.
		for ( unsigned int i = 0; i < guards.size(); ++i )
		{
			// Wait for 10 ms on each guard.
			Packet* packet = guards.at(i)->givePacket(10);

			// If we got sometring, process it.
			if ( packet )
			{
				// Initialize output list.
				packet->output = '_';

				// Prepare to check the packet. If the packet carried a
				// third-party-packet, we process it instead.
				AprsPacket* data = packet->aprsPacket;
				if ( packet->thirdParty )
				{
					data = packet->thirdParty;
				}

				// Check if we can gate this to APRS-IS.
				if ( data->canBeGatedToAprsIs() )
				{
					// Let's create APRS-IS-gateable copy of the packet.
					vector<string> addPath;
					addPath.push_back("qAR");
					addPath.push_back( string(configuration_.callsign.ToAscii()) );
					AprsPacket* tmp = new AprsPacket(*data, addPath);

					// Then we send it.
					uplink_->write(tmp);
					packet->output = 'u';
				}
				else
				{
					// Can't be gated.
					packet->flag = 't';
				}

				// Check outputting connections.
				for ( unsigned int o = 0; o < guards.size(); ++o )
				{
					if ( i != o && guards.at(o)->passesFilter(data) )
					{
						// Let's create gated-by-us copy of the packet.
						Packet* out = new Packet(*packet);
						out->inputName = string(configuration_.connections.at(i).name.ToAscii());
						out->aprsPacket = new AprsPacket(*data);
						out->thirdParty = NULL;

						// Hand it to guard.
						guards.at(o)->sendPacket(out);
					}
				}

				// Give the packet to log.
				packetlog.write(packet);
			}
			// We spent 10 ms waiting but got nothing.
			else
			{
				maintCounter++;
			}
		}


		// Check uplink for input, waiting a max of 10 ms for it.
		AprsPacket* input = uplink_->readTimeout(10);
		if ( input )
		{
			// Create packet for logging purposes.
			Packet* packet = new Packet();
			packet->timestamp = wxDateTime::Now();
			packet->input = 'u';
			packet->inputName = "TCPIP";
			packet->aprsPacket = input;
			packet->thirdParty = NULL;
			packet->output = '_';
			packet->flag = '-';

			// Check for validity.
			string errmsg;
			if ( input->error(errmsg) )
			{
				// It's invalid. Check wether it is a metadata line.
				if ( input->getTnc2Packet().at(0) != '#' )
				{
					packet->flag = 'e';
					packetlog.write(packet);
				}
				else
				{
					delete packet;
					delete input;
				}
			}
			else
			{
				// Check for duplicateness.
				if ( dupefilter.isDupe(input) )
				{
					packet->flag = 'd';
				}
				// Check if the packet can be gated.
				else if ( input->canBeGated() )
				{
					// Check outputting connections.
					for ( unsigned int o = 0; o < guards.size(); ++o )
					{
						if ( guards.at(o)->passesFilter(input) )
						{
							// Let's create gated-by-us copy of the packet.
							Packet* out = new Packet(*packet);
							out->aprsPacket = new AprsPacket(*input);
							out->thirdParty = NULL;

							// Hand it to guard.
							guards.at(o)->sendPacket(out);
						}
					}
				}
				else
				{
					packet->flag = 'n';
				}

				// Give the packet to log.
				packetlog.write(packet);
#ifdef __WXMSW__
				// Report packet to gui.
				//wxString report = packet->timestamp.Format(wxT("%Y-%m-%d %H:%M:%S"));
				//report += wxString::FromAscii(packet->aprsPacket->getTnc2Packet().c_str());
				//eventToMain(ID_newPacket, report);
#endif
			} // if (valid)
		}
		// We spent 10 ms waiting but got nothing.
		else
		{
			maintCounter++;
		}

		// Maintenance is ran at 100 ms interval.
		if ( maintCounter >= 10 )
		{
			// Keep beaconer awake.
			beacon.kick();

			// Make sure that guard dogs are awake.
			for ( unsigned int i = 0; i < guards.size(); ++i )
			{
				guards.at(i)->kick();
			}

			// Check that connections are alive.
			if ( !uplink_->isConnected() )
			{
				wxLogError(wxT("uplink has died."));
				noError = false;
				break;
			}
			for ( unsigned int i = 0; i < connections_.size(); ++i )
			{
				if ( configuration_.connections.at(i).dieOnError && !connections_.at(i)->isConnected() )
				{
					wxLogError(wxT("connection ") + wxString::Format(wxT("%d"), i) + wxT(" has died."));
					noError = false;
					break;
				}
			}

			// Maintenance done.
			maintCounter = 0;
		}
#ifndef __WXMSW__
		// Check for signals.
		if ( gotSignal_ == SIGTERM || gotSignal_ == SIGTERM )
		{
			// We need to quit.
			break;
		}
		else if ( gotSignal_ == SIGHUP )
		{
			// They want us to reopen files.
			wxLogMessage(wxT("SIGHUP received, reopening logs."));
			if ( !packetlog.reopen() || !reopenLog() )
			{
				noError = false;
			}
			else
			{
				wxLogMessage(wxT("Back in business."));
			}
			gotSignal_ = 0;
		}
#endif
	}

	// Put quitting reason in log.
#ifndef __WXMSW__
	if ( gotSignal_ )
	{
		wxLogMessage(wxT("Received SIGINT or SIGTERM from os, quitting."));
	}
	else if ( noError )
#else
	if ( noError )
#endif
	{
		wxLogMessage(wxT("Received close request, shutting down."));
	}
	else
	{
		wxLogMessage(wxT("Fatal error, terminating."));
	}

	// Remove guards.
	for ( unsigned int i = 0; i < guards.size(); ++i )
	{
	        delete guards.at(i);
	}


	// Main loop is finished. Check for errors.
	if ( !noError )
	{
#ifdef __WXMSW__
		eventToMain(ID_fatalError);
		return NULL;
#else
		return EXIT_FAILURE;
#endif
	}

	// Successfull exit.
#ifdef __WXMSW__
		return NULL;
#else
		return EXIT_SUCCESS;
#endif

}
